
class ManPage:
    """Quarto Manual Page

    Class handling the creation of the quarto markdown (`.qmd`) manual pages.

    Args:
        name (str): Name of the function, class, or method.
        obj (function or class): The function or class to be documented.
        config (Config): Object of class :py:class:`Config <pyp2qmd.Config.Config>`.
        parent (None, str): `None` (default) if a class or function is documented.
            Used to document methods, in this case `parent` contains the name of
            the parent class as str.
          
    Returns:
        Initializes an object of this class.
    """

    def __init__(self, name, obj, config, parent = None):

        from inspect import isfunction, isclass
        from .Config import Config

        # parent (None, str): If str, this will be removed from full name.
        if not isinstance(name, str):
            raise TypeError("argument `name` must be str")
        if not isfunction(obj) and not isclass(obj):
            raise TypeError("argument `obj` must be function or class")
        if not isinstance(parent, type(None)) and not isinstance(parent, str):
            raise TypeError("argument `parent` must be None or str")
        if not isinstance(config, Config):
            raise TypeError("argument `config` must be of class `Config`")

        self._name    = name
        self._obj     = obj
        self._parent  = parent
        self._config  = config

        self._doc, self._signature, self._module = self._extract_docstring()


    def _extract_docstring(self):
        """Extract Docstring

        Helper function to extract the docstring.

        Returns:
            list: Returns a list with three elements containing
                the docstring itself, the signature, and the module name.
                If no docstring is present, a list witht three `None`s is returned.
        """
        import inspect
        from docstring_parser import DocstringStyle, parse
        dstyle = getattr(DocstringStyle, self.config_get("docstringstyle").upper())

        # If parent is None, extract docstring of main function or class.
        docstring = parse(inspect.getdoc(self._obj), dstyle)
        if docstring:
            res = [docstring, inspect.signature(self._obj), self._obj.__module__]
        else:
            res = [None, None, None]

        return res

    def config_get(self, what):
        """Get Config Argument

        Interfaces the `.get()` method of the
        :py:class:`Config <pyp2qmd.Config.Config>` object.

        Args:
            what (str): Name of the attribute.

        Returns:
            Whatever is stored on the attribute.
        """
        return self._config.get(what)


    def fullname(self):
        from re import match
        if match(f"^{self._module}", self._name):
            return self._name
        else:
            return f"{self._module}.{self._name}"


    def quartofile(self):
        return f"{self.fullname()}.qmd"


    def isclass(self):
        from inspect import isclass
        return isclass(self._obj)


    def isfunction(self):
        from inspect import isfunction
        return isfunction(self._obj)


    def _format_signature(self, name, max_length = 200, remove_self = False):
        """
        formatting of signature to get some line breaks in output
        """

        from re import compile
        from html import escape

        n = max_length - len(name) - 1
        formatted_params = []
        tmp = []

        # Replacing '<lambda> at [memory pointer]>' with <lambda>
        relambda = compile("<lambda>\s+at\s+\w+>")

        for k,p in self._signature.parameters.items():
            # Removing lambda function mem addr, excaping html chars
            p = escape(relambda.sub(r"<lambda>>", str(p)))
            if remove_self and k == "self": continue
            tmp_len = max(0, sum([len(x) for x in tmp]) + (len(tmp) - 1) * 2)
            if (tmp_len + len(p) + 1) <= n:
                tmp.append(p)
            else:
                formatted_params.append(", ".join(tmp))  
                tmp = [p]

        # First empty? Can happen if max_length very small
        if len(formatted_params) > 0 and formatted_params[0] == "":
            formatted_params = formatted_params[1:]
        # End   
        if len(tmp) > 0: formatted_params.append(", ".join(tmp))
            
        spacers = "".join([" "] * (len(name) + 1))
        return name + "(" + f",<br/>{spacers}".join(formatted_params) + ")"   


    def signature(self, remove_self = None, max_length = 200):
        from re import sub

        assert isinstance(remove_self, type(None)) or isinstance(remove_self, bool)
        assert isinstance(max_length, int)
        assert max_length >= 0
        name = self._name
        # Remove ^parent. if self._parent is set
        if isinstance(self._parent, str):
            name = sub(f"^{self._parent}\.", "", self._name)
        else:
            name = self._name

        if remove_self is None:
            remove_self = self._parent is not None
        
        return self._format_signature(name, max_length, remove_self)


    def getmembers(self):
        import inspect
        members = []
        for rec in inspect.getmembers(self._obj):
            # requires three independent ifs here
            if rec[0].startswith("__"): continue
            if not inspect.isfunction(rec[1]) and not inspect.isclass(rec[1]): continue
            if not self.config_get("include_hidden") and rec[1].__name__.startswith("_"): continue
            members.append((f"{self.fullname()}.{rec[0]}", rec[1]))
        return members


    def get(self, attr):
        assert isinstance(attr, str), "argument `attr` must be of type str"
        if not hasattr(self._doc, attr):    return None
        else:                               return getattr(self._doc, attr)


    def write_qmd(self):

        from os.path import isfile
        import tempfile
        import filecmp

        qmd   = f"{self.config_get('man_dir')}/{self.quartofile()}"
        ofile = f"{self.config_get('quarto_dir')}/{qmd}"

        if not isfile(ofile):
            with open(ofile, "w+") as fid: print(self, file = fid)
        else:
            tmpfile = tempfile.NamedTemporaryFile()
            with open(tmpfile.name, "w+") as fid: print(self, file = fid)
            files_equal = filecmp.cmp(tmpfile.name, ofile)
            # New file differs? Overwrite existing qmd. Re-write the file (not
            # move the temporary file) to ensure correct file permissions.
            if not files_equal:
                with open(ofile, "w+") as fid: print(self, file = fid)
            tmpfile.close()
        
        # Return name of the qmd; used for linking
        return qmd


    def __repr_args(self):

        from re import findall, sub
        from html import escape

        res = ""

        # Given the signature, we should have the following parameters
        expected_args = list(self._signature.parameters.keys())
        # These are the available parameters (from the docstring)
        documented_args = [sub("^\*+", "", x.arg_name.strip()) for x in self.get("params")]
        # If self._parent is set this is the man page for a method.
        # if the first argument in list is 'self', remove.
        if self._parent and len(expected_args) > 0 and expected_args[0] == "self":
            expected_args.remove("self")

        # Check if any of the expected arguments is not documented
        missing_args = []
        for ea in expected_args:
            if not ea in documented_args:
                missing_args.append(ea)

        # Counting number of warnings and/or arguments added
        counter = 0

        if len(missing_args) > 0:
            counter += 1
            res += "<ul>"
            for rec in missing_args:
                res += f"<li>WARNING(missing argument definition \"{rec}\" in docstring)</li>"
            res += "</ul>"

        # Count arguments
        res += "<dl class=\"pyp-list param-list\">\n"
        for arg in self.get("params"):
            counter += 1
            # If we get "argument (class)" we separate them
            mtch = findall(r"^(.*)\((.*?)\)$", arg.args[1])
            if len(mtch) > 0:
                arg_name = mtch[0][0].strip()
                arg_cls  = f"<code class=\"argument-class\">{escape(mtch[0][1])}</code>"
            else:
                arg_name = arg.args[1].strip()
                arg_cls  = ""

            #short_arg = re.search("^([\*\w]+)", arg.args[1]).group(1).replace("*", "")
            # Building html table row
            res += "  <dt style = \"white-space: nowrap; font-family: monospace; vertical-align: top\">\n" + \
                   f"   <code id=\"{self.fullname()}:{arg_name}\">{escape(arg_name)}</code>{arg_cls}\n" + \
                   "  </dt>\n" + \
                   f" <dd>{self._add_references(arg.description)}</dd>\n"
                   ####f" <dd>{escape(self._add_references(arg.description))}</dd>\n"

        res += "</dl>"

        # If no arguments or warnings have been added, return None
        # such that we can skip to add the 'Arguments' section.
        return None if counter == 0 else res


    def __repr_raises(self):
        import xml.etree.ElementTree as et
        from html import escape
        res = "<ul class=\"python-raises\">\n"
        for rec in self.get("raises"):
            res += f"<li><code class=\"text-warning\">{rec.type_name}</code>: " + \
                   f"{escape(self._add_references(rec.description))}\n"
        return res + "</ul>\n"


    def __repr__(self):
        import re

        if self.get("short_description") is None:
            res  = "---\ntitle: \"WARNING(short_description missing)\"\n---\n\n"
        else:
            res  = "---\ntitle: \"" + \
                   self._add_references(self.get("short_description")) + \
                   "\"\n---\n\n"

        if self.get("long_description"):
            res += "### Description\n\n"
            res += self._add_references(self.get("long_description"))
        else:
            res += "WARNING(long_description missing)"

        res += "\n\n### Usage\n\n"
        res += "<pre><code class='language-python'>" + \
               f"{self.signature(max_length = 50)}" + \
               "</code></pre>"

        # Function arguments
        tmp = self.__repr_args()
        if tmp is not None:
            res += "\n\n### Arguments\n\n" + tmp

        # Return value
        if self.get("returns"):
            res += "\n\n### Return\n\n"
            res += f"{self._add_references(self.get('returns').description)}"

        # If is class, append methods
        if self.isclass():

            res += "\n\n### Methods\n\n"
            # Convert package.module.class into package.module
            parent = re.sub(r"\.[^.]*$", "", self.fullname())

            res += "<dl class=\"pyp-list method-list\">\n"
            for name,meth in self.getmembers():
                m_man = ManPage(name, meth, self._config)
                if m_man.get("short_description") is None:
                    short = "WARNING(short_description missing)"
                else:
                    short = m_man.get("short_description")

                # Adding <dt><dd> for current method
                link = m_man.quartofile()
                text = re.sub(f"^{parent}\.", "", m_man.signature(remove_self = True))
                res += "    <dt style = \"white-space: nowrap; font-family: monospace; vertical-align: top\">\n" + \
                       f"       <code>[{text}]({link})</code>\n    </dt>\n" + \
                       f"    <dd>{short}</dd>\n"
            res += "</dl>\n"

 
        # If we have examples:
        if self.get("examples"):
            examples = []
            for tmp in [ex.description for ex in self.get("examples")]:
                # Prepare example and possibly split it.
                tmp       = self._prepare_example(tmp)
                examples += self._split_example(tmp)

            res += "\n\n### Examples\n\n"
            for tmp in examples:
                res += self.__repr_examples(tmp)
            res += "\n"

        # If has documented raises exception
        if len(self.get("raises")) > 0:
            res += "\n\n### Raises\n\n" + self.__repr_raises()

 
        return res

    def _add_references(self, x):
        import re
        if x is None: return x

        # Replace sphinx refs with quarto refs
        matches = re.findall(r":py:(?:func|class|method):`.*?`", x)

        # If any matches are found, set quarto references properly
        for m in matches:
            # Find basic match
            tmp = re.search("py:(\w+?):`(.*?)(?=`)", m)
            if tmp:
                # Extact typ (func, class, or method) and the 'reference'
                typ,ref = tmp.groups()
                # If format is "name <ref>" we further decompose the match
                ref2 = re.match("^(.*)\s+?<(.*?)>$", ref)
                # Take `text <link>` from the docstring
                if ref2:
                    x = x.replace(m, f"[{ref2.group(1)}]({ref2.group(2)}.qmd)")
                # IF we only have a `word` we expect that it refers
                # to it's current module OR its class (if typ == "method")
                elif re.match("^\w+$", ref):
                    if typ == "method":
                        cls = re.sub(r"\.\w+?$", "", self.fullname())
                        x   = x.replace(m, f"[{ref}]({cls}.{ref}.qmd)")
                    else:
                        x   = x.replace(m, f"[{ref}]({self._obj.__module__}.{ref}.qmd)")
                # Else take whatever we got
                else:
                    x = x.replace(m, f"[{ref}]({ref}.qmd)")

        return x


    def __repr_examples(self, x):
        from re import sub, MULTILINE
        assert isinstance(x, str)
        res = "```{python}\n" + \
              "#| echo: true\n" + \
              "#| error: true\n" + \
              "#| warning: true\n" + \
              sub("^#:", "#", x, flags = MULTILINE) + \
              "\n```\n\n"
        return res


    def _prepare_example(self, x):
        """prepapre_example(x)
 
        Args:
            x (str): The example extracted from the docstring.
 
        Return:
        str : Modified example to be ready for quarto.
        """
        from re import sub, MULTILINE
        # Comment 'text' (can be included in the py example section)
        x = sub(r"^(?!(>>>))", "## ", x, flags = MULTILINE)
        # Removing empty lines
        x = sub(r"^##\s+?$", "", x, flags = MULTILINE)
        # Remove >>> code identifiers
        x = sub(r"^>>>\s+?", "", x, flags = MULTILINE)
        return x


    def _split_example(self, x):
        """split_example(x)
 
        Args:
            x: str, the example extracted from the docstring.
 
        Return:
        list: List of strings. If we find `#:` at the start of a line
        we split the example at this position in multiple segments.
        """
        from re import split, sub, MULTILINE
        res = split(r"\n(?=#:)", x, flags = MULTILINE)
        # Remove empty lines if any
        for i in range(len(res)):
            res[i] = sub("#:[\s+]?\n", "", res[i], flags = MULTILINE)
        return res








# Class to handle the setup/config
from .Config import Config

# Main class handling the conversion (docstrings to quarto markdown)
from .DocConverter import DocConverter

# Manual page handler
from .ManPage import ManPage

# Series of functions for testing only
from .demofunctions import *



#!/usr/bin/env python3
# -------------------------------------------------
# Building quarto manual pages for python package
# -------------------------------------------------


class DocConverter:
    """Documentation Converter

    This is the main class of the package which will extract the docstrings
    of all exported classes and functions of an installed python package,
    and convert them into quarto markdown files (thus pyp2qmd; python package to
    quarto markdown).

    The :py:class:`Config <pyp2qmd.Config.Config>` contains all the required
    arguments/settings to perform this task and is, thus, the only reqired
    input argument.

    Args:
        config (Config): See :py:class:`Config <pyp2qmd.Config.Config>` for details; must be
            set up properly.

    Return:
        Initializes a new object of class `DocConverter`.

    Raises:
        TypeError: If `config` is not of class :py:class:`Config <pyp2qmd.Config.Config>`.
        Exception: If `config` object is not set up properly.
        Exception: If the package which should be documented is not found (can't
            be imported).
        ValueError: If package `docstring_parser` has no `DocstringStyle` class
            matching the docstringstyle defined by the user (handled as 'all upper case').
    """

    # bool: is used to later on modify _quarto.yml (adding function references
    # and class references).
    _quarto_yml_initialized = False

    # Created man pages will be stored here, used to populate _quarto.yml if needed
    _man_created = {"class": dict(), "function": dict(), "method": dict()}

    def __init__(self, config):
        from .Config import Config
        from importlib import import_module
        from docstring_parser import DocstringStyle
        if not isinstance(config, Config):
            raise TypeError("argument `config` must be of class pyp2qmd.Config")
        elif not config.is_set_up():
            raise Exception(f"issue with argument `config`: {config}")

        # Testing two more things. First we load the package which should be
        # documented. Throws an error if that package cannot be loaded (not
        # installed). Stores the package on _pkg, used by the methods to extract
        # classes and functions (e.g., get_classes, get_functions methods).
        try:
            self._pkg = import_module(config.get("package"))
        except Exception as e:
            raise Exception(f"Cannot document \"{config.get('package')}\". Reason: {e}.")

        # Next check if the docstringstyle is valid
        if not hasattr(DocstringStyle, config.get("docstringstyle").upper()):
            raise ValueError("invalid `docstringstyle`, package `docstring_parser` has no class " + \
                           f"`docstring_parser.DocstringStyle.{config.get('docstringstyle').upper()}`")

        # Store config
        self._config = config

        # Checking action: If action = "init" and overwrite = False we are
        # checking if some specific output files already exist. If so, raise
        # Exception and inform the user that he/she can enable overwrite,
        # however, this might result in loss of data (overwrites these files).
        from os.path import join, isfile
        files_to_check = [
                join(self.config_get("quarto_dir"), "_quarto.yml"),
                join(self.config_get("quarto_dir"), "pyp.scss"),
                join(self.config_get("quarto_dir"), "index.qmd")
            ]
        if self.config_get("action") == "init" and not self.config_get("overwrite"):
            tmp = []
            for f in files_to_check:
                if isfile(f): tmp.append(f)

            if len(tmp) > 0:
                pl  = "" if len(tmp) == 1 else "s"
                pl2 = "it" if len(tmp) == 1 else "them"
                tmp = ", ".join(f"\"{x}\"" for x in tmp)
                raise Exception(f"file{pl} {tmp} already exist{pl}, `action = \"init\"` " + \
                        f"would overwrite {pl2} with a template. Consider using " + \
                        "`action = \"document\"` or allow DocConverter to overwrite " + \
                        "the existing file using `overwrite = True` (`--overwrite` " + \
                        "via console). WARNING: May `action = \"init\"` with " + \
                        f"`overwrite = True` may result in loss of data if the file{pl} " + \
                        "mentioned have been edited manually!")

        # Proceed. If action == 'init' create template files
        if self.config_get("action") == "init":
            self.__init_documentation()
            self._quarto_yml_initialized = True 

        # Class is now ready to do what it is designed for

    def __init_documentation(self):

        from os import makedirs
        from os.path import join, isdir, basename
        from shutil import copy
        from re import sub, match
        
        # Trying to create quarto_dir and quarto_dir/man_dir if needed
        if not isdir(self.config_get("quarto_dir")):
            try:
                makedirs(self.config_get("quarto_dir"))
            except Exception as e:
                raise Exception(f"cannot create {self.config_get('quarto_dir')}: {e}")
        man_dir = join(self.config_get("quarto_dir"), self.config_get("man_dir"))
        if not isdir(man_dir):
            try:
                makedirs(man_dir)
            except Exception as e:
                raise Exception(f"cannot create {man_dir}: {e}")
         
        # Getting package name
        pkgname = match(r"^([^\.]+)", self.__class__.__module__).group(1)

        # Adding template(s)
        def pkg_file(pkgname, directory, file):
            from re import match
            from pkg_resources import resource_filename
            from os.path import isfile, join

            # Getting package name
            assert isinstance(pkgname, str), TypeError("argument `pkgname` must be str")
            assert isinstance(file, str), TypeError("argument `file` must be str")
            assert isinstance(directory, str), TypeError("argument `directory` must be str")

            # Getting name of file
            file = resource_filename(pkgname, join(directory, file))
            if not isfile(file):
                raise Exception(f"whoops, file \"{file}\" (intended to be shipped with " + \
                        "the package) does not exist; contact the mainainer")
            return file

        # Adding _quarto.yml
        src = pkg_file(pkgname, "templates", "_quarto.yml")
        content = open(src, "r").read()
        content = sub("<title>", pkgname, content)
        content = sub("<output_dir>", self.config_get("output_dir"), content)
        with open(join(self.config_get("quarto_dir"), basename(src)), "w") as fid:
            fid.write(content)
        del src, content

        # Adding index.qmd
        from datetime import datetime as dt
        src = pkg_file(pkgname, "templates", "index.qmd")
        content = open(src, "r").read()
        content = sub("<title>", pkgname, content)
        content = sub("<date_and_time>", f"{dt.now():%Y-%m-%d %H:%M}", content)
        with open(join(self.config_get("quarto_dir"), basename(src)), "w") as fid:
            fid.write(content)
        del src, content

        # Copy scss
        src = pkg_file(pkgname, "templates", "pyp.scss")
        copy(src, join(self.config_get("quarto_dir"), "pyp.scss"))




    def __repr__(self):
        """Standard Representation

        Standard representation of the DocConverter class,
        mainly for development purposes at the moment.

        Returns:
            str: String with standard representation.
        """
        res  = f"{self.__module__} Object:\n"

        cls = self.get_classes(names_only = True)
        if len(cls) > 0:
            res += f"{self.__nice_stdout_list_of_strings('    Classes:   ', cls)}"

        fun = self.get_functions(names_only = True)
        if len(fun) > 0:
            res += f"{self.__nice_stdout_list_of_strings('    Functions: ', fun)}"

        return res


    def __nice_stdout_list_of_strings(self, desc, x):
        from re import sub

        assert isinstance(desc, str), TypeError("argument `desc` must be str")
        assert isinstance(x, list),   TypeError("argument `x` must be str")
        assert all([isinstance(y, str) for y in x]), TypeError("elements in `x` must be str")

        res = [desc]
        for rec in x:
            tmp = res[len(res) - 1] + rec + ", "
            if len(tmp) < 80:
                res[len(res) - 1] = tmp
            else:
                res.append(" " * len(desc) + rec + ", ")
        return sub(",\s+$", "", "\n".join(res)) + "\n"


    def config_get(self, what):
        """Get Config Attribute

        Calls the `.get()` method of the
        :py:class:`Config <pyp2qmd.Config.Config>` object
        as defined via `config` on class initialization. Will raise
        exceptions if not available (see :py:class:`Config <pyp2qmd.Config.Config>`
        for details).

        Args:
            what (str): Attribute to be returned.

        Return:
            Whatever is stored on the `_{what}` attribute of the
            :py:class:`Config <pyp2qmd.Config.Config>` object.
        """
        return self._config.get(what)


    def get_classes(self, names_only = False):
        """Get Exported Classes

        Args:
            names_only (bool): If `False` (default) a list of tuples is returned,
                if set `True` a list of str (name of classes only).

        Return:
            dict or list: Dictionary with all exported classes if `names_only = True`,
                else a list of str containing only the class names.

        Raises:
            Exception: If there are issues extracting classes via `inspect`.
        """
        from inspect import getmembers, isclass
        try:
            res = getmembers(self._pkg, isclass)
        except Exception as e:
            raise Exception(f"problems extracting classes from package: {e}")

        if not names_only:
            return dict(zip([x[0] for x in res], [x[1] for x in res]))
        else:
            return [x[0] for x in res]

    def get_functions(self, names_only = False):
        """Get Exported Functions

        Args:
            names_only (bool): If `False` (default) a list of tuples is returned,
                if set `True` a list of str (name of classes only).

        Return:
            dict or list: Dictionary with all exported functions if `names_only = True`,
                else a list of str containing only the function names.

        Raises:
            Exception: If there are issues extracting functions via `inspect`.
        """
        from inspect import getmembers, isfunction
        try:
            res = getmembers(self._pkg, isfunction)
        except Exception as e:
            raise Exception(f"problems extracting classes from package: {e}")

        if not names_only:
            return dict(zip([x[0] for x in res], [x[1] for x in res]))
        else:
            return [x[0] for x in res]


    def document_functions(self):
        """Document Functions

        Generates man pages for all exported functions.
        """
        from .ManPage import ManPage

        for name,cls in self.get_functions().items():
            if not self.config_get("silent"):
                print(f"Create man page for function {name}")
            man = ManPage(name, cls, self._config)
            self._man_created["function"][name] = man.write_qmd()
    

    def document_classes(self):
        """Document Classes

        Generates man pages for all exported classes.
        """
        from .ManPage import ManPage
        from re import sub

        for name,cls in self.get_classes().items():
            if not self.config_get("silent"):
                print(f"Create man page for class {name}")
            man = ManPage(name, cls, self._config)
            self._man_created["class"][name] = man.write_qmd()

            for name,meth in man.getmembers():
                if not self.config_get("include_hidden") and meth.__name__.startswith("_"):
                    continue
                if not self.config_get("silent"):
                    print(f"   + method page for {name}")
                parent = sub(r"\.[^.]*$", "", man.fullname())
                m_man = ManPage(name, meth, self._config, parent = parent)
                self._man_created["method"][name] = m_man.write_qmd()


    def document(self):
        """Document All

        Documents all exported classes and functions. Convenience function,
        calls :py:meth:`document_functions` and :py:meth:`document_classes`.
        """
        self.document_functions()
        self.document_classes()

    
    def update_quarto_yml(self):
        """Update Quarto

        Updates the `_quarto.yml` file by adding all classes and functions
        to the website sidebar content (navigation). This is only done if
        the quarto file has just been initialized and man pages have been
        crated.
        """

        n = sum([len(x) for x in self._man_created])
        if not self.config_get("silent"):
            print(f"pyp2qmd: Number of man pages created {n}")
        # Nothing? Do nothing
        if not self._quarto_yml_initialized or n == 0:
            return

        # Reading existing yml file
        content = self._load_yaml()
            
        content["website"]["sidebar"]["contents"] = []
        # Setting up dictionary for function references
        for what in ["Function", "Class"]:
            if len(self._man_created[what.lower()]):
                tmp = []
                for key,val in self._man_created[what.lower()].items():
                    tmp.append({"text": key, "file": val})
                tmp = {"section": f"{what} references", "contents": tmp}
                content["website"]["sidebar"]["contents"].append(tmp)

        # Write back
        self._save_yaml(content)


    def add_navbar_page(self, src, dest, text, menu = None):
        """Add Page to Navigation

        Adds page to website navbar left. Will be added to the _quarto.yml
        file if not yet included. Will place the source file (qmd) in the
        quarto output folder under name `dest` and linked in the navigation
        as `text`.

        Args:
            src (str): Path to an existing quarto file, must end in `.qmd`.
            dest (str): Name or path for target quarto file. If path, the
                directory will be created inside `quarto_dir` as specified in the
                config (see :py:class:`DocConverter`) if not yet existing.
            text (str): Name used in the navigation.
            menu (None, str): Must be `None` if pages are added. If set,
                it is expected that a menu with this name exists. The page
                will then be added to that menu if not already in there.

        Raises:
            TypeError: If `src`, `dest`, `text` are not str.
            ValueError: If `src` and `dest` do not end in `.qmd`
            ValueError: If `dest` is a path, not only the name of the target quarto file.
            FileNotFoundError: If `src` does not exist.
            TypeError: If `menu` is not `None` nor `str`.
            Exception: If `dest` already exists but overwrite is set `False`
                (see :py:class:`DocConverter`).
            Exception: if `src` cannot be copied to destination.
        """


        from re import match
        from os.path import isfile, basename, dirname, join
        from os import makedirs
        from shutil import copy

        if not isinstance(src, str):
            raise TypeError("argument `src` must be str, qmd file name")
        elif not match(r".*\.qmd$", src):
            raise ValueError("argument `src` must end in '.qmd' pointing to a quarto file")
        elif not isfile(src):
            raise FileNotFoundError(f"cannot find file \"{src}\"")

        if not isinstance(dest, str):
            raise TypeError("argument `dest` must be str, qmd file name")
        elif not match(r".*\.qmd$", dest):
            raise ValueError("argument `dest` must end in '.qmd' pointing to a quarto file")

        # Path?
        if not dest == basename(dest):
            tmp = join(self.config_get("quarto_dir"), dirname(dest))
            try:
                makedirs(tmp, exist_ok = True)
            except Exception as e:
                raise Exception(f"cannot create folder \"{tmp}\": {e}")

        if not isinstance(text, str):
            raise TypeError("argument `text` must be str, link name")

        if not isinstance(menu, type(None)) and not isinstance(menu, str):
            raise TypeError("argument `menu` must be `None` or str")

        # Already exists?
        dest_path = join(self.config_get("quarto_dir"), dest)
        if isfile(dest_path) and not self.config_get("overwrite"):
            raise Exception(f"target file \"{dest_path}\" already exists and overwrite " + \
                    "is False. Consider enabling overwrite. Could result in loss of data!")

        # Reading existing yml file
        content = self._load_yaml()

        # Access existing navbar left. If this fails, it does not exist
        # in the _quarto.yml and an exception will be thrown.
        try:
            tmp = content["website"]["navbar"]["left"]
        except:
            raise Exception(f"\"{ymlfile}\" does not contain website > navbar > left")

        # Copy src file to dest_path using shutil
        try:
            copy(src, dest_path)
        except Exception as e:
            raise Exception(e)

        # If `menu = None` we are adding a page, so we are searching
        # (and appending) to  this:
        nav = content["website"]["navbar"]["left"] # 'pointer'
        # If `menu` is set, we must first find the menu.
        if isinstance(menu, str):
            found = False
            for j in range(len(nav)):
                if not "menu" in nav[j].keys() or not "text" in nav[j].keys():
                    continue
                elif nav[j]["text"] == menu:
                    # Replace `nav` object and break
                    nav = nav[j]["menu"]
                    found = True
                    break
            # Not found? Error
            if not found:
                raise Exception(f"could not find menu \"{menu}\". " + \
                                "Not yet added via .add_navbar_menu()?")


        # Loop trough existing entries. If we find an entry
        # with the current file text, just update the text. Else add at the end.
        # This is when adding pages directly to the navigation.
        found = False
        for i in range(len(nav)):
            if len(nav[i]) > 0 and nav[i]["file"] == dest:
                nav[i]["text"] = text
                found = True
                break

        # Not found? Append at the end.
        if not found: nav.append({"file": dest, "text": text})

        # Write back
        self._save_yaml(content)


    def add_navbar_right(self, x):
        """Add Element To Navbar Right

        Allows to add elements to website navbar right.
        Will add it if not yet existing. Accepts a dict, originally
        used to add repository logo and link.
        """
        if not isinstance(x, dict):
            raise TypeError("argument `x` is expected to be a dict")

        content = self._load_yaml()
        try:
            tmp = content["website"]["navbar"]
        except:
            raise Exception("_quarto.yml does not contain website > navbar")

        try:
            tmp = content["website"]["navbar"]["right"]
        except:
            # Initialize empty list element
            tmp["right"] = []
            tmp = tmp["right"]
        tmp.append(x)

        self._save_yaml(content)

    def add_navbar_menu(self, menu):
        """Add Dropdown Menu to Navigation

        Adds a menu (dropdown menu) to the top navigation which can be
        populated with a series of pages. Will only be added if not yet existing.

        Args:
            menu (str): Name of the dropdown menu.

        Raises:
            TypeError: If `menu` is not str.
        """


        from re import match
        from os.path import isfile, basename

        if not isinstance(menu, str):
            raise TypeError("argument `menu` must be str")

        # Reading existing yml file
        content = self._load_yaml()

        # Check for menus
        found = False
        for rec in content["website"]["navbar"]["left"]:
            # No menu? Or no text? Skip
            if not "menu" in rec.keys() or not "text" in rec.keys():
                continue
            # Is the current menu the one the user wants to add?
            if rec["text"] == menu:
                found = True
        
        # Not found: Add new empty menu
        if not found:
            tmp = {"text": menu, "menu": []}
            content["website"]["navbar"]["left"].append(tmp)

        # Write back
        self._save_yaml(content)


    def add_favicon(self, file):
        """Add Favicon

        Args:
            file (str): File name of parth to the favicon (png, jpg).

        Raises:
            TypeError: If `file` is not str.
            FileNotFoundError: If `file` does not point to an existing file.
        """

        from re import match
        from os.path import isfile, basename, join
        from shutil import copy

        if not isinstance(file, str):
            raise TypeError("argument `file` must be str")
        elif not isfile(file):
            raise FileNotFoundError(f"file \"{f}\" not found on disc")

        # Reading existing yml file
        content = self._load_yaml()

        # Try to find the 'website' section where we will add this option
        content["website"]["favicon"] = basename(file)

        # Write back
        self._save_yaml(content)

        # Copy src file to dest_path using shutil
        try:
            copy(file, join(self.config_get("quarto_dir"), basename(file)))
        except Exception as e:
            raise Exception(e)

    def add_logo(self, logo, title):
        """Adding Logo and/or Title

        Allows to add a logo image and/or change the title.

        Args:
            logo (str, None): String (image file name) or None. If None,
                the logo is not set or removed if needed.
            title (str, None): String (title) or None. If None, title is set false
                such that it is not shown in navbar.
        """
        if not isinstance(logo, (str, type(None))):
            raise TypeError("argument `logo` must be str or None")
        if not isinstance(title, (str, type(None))):
            raise TypeError("argument `title` must be str or None")

        content = self._load_yaml()
        try:
            tmp = content["website"]["navbar"]
        except:
            raise Exception("_quarto.yml does not contain expected website > navbar")

        if isinstance(title, type(None)):
            tmp["title"] = False
        else:
            tmp["title"] = title

        if isinstance(logo, str):
            tmp["logo"] = logo
        elif "logo" in tmp.keys():
            del tmp["logo"]

        self._save_yaml(content)


    def _load_yaml(self):
        """Load Existing YML File

        Loads the existing _quarto.yml file.
        """
        from os.path import join
        import yaml
        ymlfile = join(self.config_get('quarto_dir'), "_quarto.yml")
        with open(ymlfile, "r") as fid:
            content = yaml.load("".join(fid.readlines()), yaml.SafeLoader)
        return content


    def _save_yaml(self, content):
        """Save (Updated) YML File

        Writes the dictionary back to `_quarto.yml`.
        """
        from os.path import join
        import yaml
        assert isinstance(content, dict), TypeError("argument `content` expected to be dict")
        ymlfile = join(self.config_get('quarto_dir'), "_quarto.yml")
        with open(ymlfile, "w+") as fid: fid.write(yaml.dump(content))


    def _add_website_option(self, key, value):
        assert isinstance(key, str), TypeError("argument `key` must be str")
        assert isinstance(value, str), TypeError("argument `value` must be str")

        content = self._load_yaml()

        # Access existing navbar left. If this fails, it does not exist
        # in the _quarto.yml and an exception will be thrown.
        try:
            tmp = content["website"]
        except:
            raise Exception(f"\"{ymlfile}\" does not contain website > navbar > left")

        tmp[key] = value
        self._save_yaml(content)

    def add_repo_url(self, url, branch = "main"):
        """Adding Soruce Code Repository URL

        Args:
            url (str): URL (e.g., to the github repository).
            branch (str): Branch name, defaults to "main".

        Raises:
            TypeError: If `url` or `branch` are not str.
            ValueError: If `url` does not look like a URL.
        """

        from re import match

        if not isinstance(url, str): raise TypeError("argument `url` must be str")
        if not isinstance(branch, str): raise TypeError("argument `url` must be str")
        if not match("^https?:\/\/", url):
            raise ValueError(f"url (\"{url}\") does not look like a valid URL")

        self._add_website_option("repo-url", url)
        self._add_website_option("repo-branch", branch)

    def add_issue_url(self, url):
        """Adding URL to Issue Trackign Page

        Args:
            url (str): URL (e.g., to the github issues page).

        Raises:
            TypeError: If `url` is not str.
            ValueError: If `url` does not look like a URL.
        """

        from re import match

        if not isinstance(url, str): raise TypeError("argument `url` must be str")
        if not match("^https?:\/\/", url):
            raise ValueError(f"url (\"{url}\") does not look like a valid URL")

        self._add_website_option("issue-url", url)


    def add_scss(self, file):
        from os.path import isfile, join
        if not isinstance(file, str):
            raise TypeError("argument `file` must be str")
        elif not isfile(join(self.config_get("quarto_dir"), file)):
            raise FileNotFoundError(f"missing \"{join(self.config_get('quarto_dir'), file)}\"")

        content = self._load_yaml()

        # Access existing navbar left. If this fails, it does not exist
        # in the _quarto.yml and an exception will be thrown.
        try:
            tmp = content["format"]["html"]["theme"]
        except:
            raise Exception(f"\"{ymlfile}\" does not contain format > html > theme")

        # Append in second position
        if not file in tmp:
            if not isinstance(tmp, list) or len(tmp) == 0:
                content["format"]["html"]["theme"] = [file]
            else:
                content["format"]["html"]["theme"] = [tmp[0], file] + tmp[1:]

        self._save_yaml(content)








class Config:
    """pyp2qmd Configuration

    Custom class for handling the required arguments for pyp2qmd which
    has two modes: If `argparse = False` (default) on initialization, the
    user must use the :py:meth:`setup` method to specify the required
    arguments.

    If `argparse = True` user inputs will be parsed via the `argparse` package
    and automatically handed over to the :py:meth:`setup` method. This is used
    when called via console (console arguments).

    Args:
        argparse (bool): `False` by default, the user has to use the `.setup()` method.
            If set to `True` it will try to parse input arguments via `argparse`.

    Return:
        Initializes an object of class `pyp2qmd.Configuration`.
    """

    def __init__(self, argparse = False):

        # Falls until .setup was executed such that we know
        # that the object has been set up properly one way
        # or the other.
        self._is_set_up = False

        # Parse user input args
        if argparse: self.__parse_arguments()

    def __parse_arguments(self):

        # Allowed action options
        allowed_action = ["init", "document"]

        import argparse
        from re import search
        pkg_name = search(r"^(.*?)(?=\.)", self.__module__).group(1)
        parser = argparse.ArgumentParser(pkg_name)
        parser.add_argument("action", nargs = 1, type = str,
                help = f"Action to perform, one of: {', '.join(allowed_action)}")
        parser.add_argument("-p", "--package", type = str,
                help = "Name of the python package.")
        parser.add_argument("--overwrite", default = False, action = "store_true",
                help = "Only used if action = create; will overwrite _quarto.yml if needed.")
        parser.add_argument("--quarto_dir", type = str, default = "_quarto",
                help = "Name of the target directory for rendered quarto website. " + \
                       "Defaults to \"_quarto\".")
        parser.add_argument("--man_dir", type = str, default = "man",
                help = "Folder to store the function/class manual qmds (subfolder of quarto_dir). " + \
                       "Defaults to \"man\".")
        parser.add_argument("--output_dir", type = str, default = "_site",
                help = "Name of the target directory for rendered quarto website," + \
                       "relative to quarto_dir. Defaults to \"_site\".")
        parser.add_argument("--docstringstyle", type = str, default = "GOOGLE",
                help = "Docstring type (format). Defaults to \"GOOGLE\".")
        parser.add_argument("--include_hidden", default = False, action = "store_true",
                help = "If set, hidden functions and methods will also be documented " + \
                        "(functions/methods starting with _ or __).")
        parser.add_argument("--silent", default = False, action = "store_true",
                help = "If set, output will be suppressed.")

        # Parsing input args
        args = parser.parse_args()

        args.action = args.action[0]
        if not args.action in allowed_action:
            parser.print_help()
            sys.exit("\nUsage error: invalid \"action\" (see help).")

        if args.package is None:
            parser.print_help()
            sys.exit("\nUsage error: argument -p/--package must be set.")

        ymlfile = f"{args.quarto_dir}/_quarto.yml"
        if args.action == "create" and isfile(ymlfile):
            parser.print_help()
            sys.exit(f"\nUsage error: action is set to \"{args.action}\" " + \
                     f"but the file \"{ymlfile}\" " + \
                     f"already exists. Remove folder \"{args.quarto_dir}\" or use " + \
                     f"--overwrite; be aware, will overwrite the existing \"{ymlfile}\".")


        self.setup(**vars(args))


    def setup(self, action, package,
              quarto_dir = "_quarto", man_dir = "man", output_dir = "_site",
              overwrite = False, include_hidden = False, docstringstyle = "GOOGLE",
              silent = False):
        """Config Setup

        `action = "init"` initializes the auto-generated documentation and will
        automatically create the quarto website template including a
        `_quarto.yml` file and `pyp.sass`. Will thorw an exception one of these
        files already exists, except if the user allows for `overwrite = True`
        (take care, the current content of `_quarto.yml` will be overwritten).

        `action = "document"` parses the python package classes and functions
        and will update the man pages. It will, however, not overwrite
        `_quarto.yml`. When used the first time, it will also initialize the
        output folder structure and create the required files (`_quarto.yml`,
        `pyp.sass`) similar to the init action.

        Args:
            action (str): Action to be executed. One of `"init"` or `"document"`,
                see method description.
            package (str): Name of the package which should be documented.
            quarto_dir (str): Output directory, defaults to `"_quarto"`.
            man_dir (str): Name of the directory for the manual pages (subfolder
                inside `quarto_dir`), defaults to `"man"`.
            output_dir (str): Directory for the rendered quarto website, used
                as `output-dir` target in `_quarto.yml`, relative to
                `quarto_dir`. Defaults to `"_site"`.
            overwrite (bool): Only used if `action = "init"`, see method
                description.
            include_hidden (bool): If `False` (default), classes, functions, and
                methods starting with an underscore will not be documented
                (no quarto man pages will be crated). Dunder classes, functions, and
                methods are always excluded.
            docstringstyle (str): Style of the docstrings in the package, must be one
                of the allowed types of the `docstring_parser` package
                (AUTO, EPYDOC, GOOGLE, NUMPYDOC, REST), defaults to `"GOOGLE"`;
                not case sensitive.
            silent (bool): If `False` (default) some output will be produced
                when rendering the man pages. Can be specified to silence the
                execution.

        Raises:
            TypeError: If the inputs are not of the expected type.
            ValueError: If `action` is not one of the allowed ones.
        """

        # Store input arguments as object attributes
        for key,val in locals().items():
            if key in ["self", "kwargs"]: continue
            setattr(self, f"_{key}", val)

        # --------------------------------------------
        # Now checking validity of all required args
        # --------------------------------------------
        action_allowed = ["init", "document"]
        if not isinstance(self.get("action"), str):
            raise TypeError("argument `action` must be str")
        elif not self.get("action") in action_allowed:
            raise ValueError(f"action must be one of: {', '.join(action_allowed)}")

        if not isinstance(self.get("package"), str):
            raise TypeError("argument `package` must be str")

        if not isinstance(self.get("quarto_dir"), str):
            raise TypeError("argument `quarto_dir` must be str")
        if not isinstance(self.get("man_dir"), str):
            raise TypeError("argument `man_dir` must be str")
        if not isinstance(self.get("output_dir"), str):
            raise TypeError("argument `output_dir` must be str")

        if not isinstance(self.get("overwrite"), bool):
            raise TypeError("argument `overwrite` must be bool")
        if not isinstance(self.get("include_hidden"), bool):
            raise TypeError("argument `include_hidden` must be bool")

        if not isinstance(self.get("docstringstyle"), str):
            raise TypeError("argument `docstringstyle` must be str")

        if not isinstance(self.get("silent"), bool):
            raise TypeError("argument `silent` must be bool")

        self._is_set_up = True


    def get(self, what):
        """Get Attribute

        Allows to access attributes from the object. `get("foo")`
        will try to return `self._foo` if it exists.

        Args:
            what (str): Name of the attribute (without leading underscore).

        Returns:
            Whatever is stored on the attribute.

        Raises:
            TypeError: If argument `what` is not str.
            ValueError: If the argument `_{what}` does not exist.
        """
        if not isinstance(what, str):
            raise TypeError("argument `what` must be str")
        if not hasattr(self, f"_{what}"):
            raise ValueError(f"object has no attribute \"_{what}\"")
        return getattr(self, f"_{what}")


    def is_set_up(self):
        """Object Set Up Properly?

        Returns:
            bool: Returns `True` if the object has been set up properly
            (see main class description) and `False` otherwise.
        """
        return self._is_set_up

    def __repr__(self):
        """Standard Representation

        Returns:
            str: Standard representation of the class.
        """
        if not self.is_set_up():
            return f"{self.__module__} Object not yet set up, call .setup() first."
        else:
            res  = f"{self.__module__} Object:\n"
            res += f"    Action:            {self.get('action')}\n"
            res += f"    Package:           {self.get('package')}\n"
            res += f"    Quarto dir:        {self.get('quarto_dir')}\n"
            res += f"    Man page dir:      {self.get('man_dir')}\n"
            res += f"    Output dir:        {self.get('output_dir')}\n"
            res += f"    Overwrite:         {self.get('overwrite')}\n"
            res += f"    Include hidden:    {self.get('include_hidden')}\n"
            res += f"    Docstring style:   {self.get('docstringstyle')}\n"
            return res






#!/usr/bin/env python

def main():
    from pyp2qmd import Config, DocConverter

    # Initialize Config; parses user arguments via argparse
    config  = Config(argparse = True)
    if not config.get("silent"): print(config)

    # Initialize DocConverter; creates _quarto.yml,
    # pyp.sass, and index.qmd if needed.
    docconv = DocConverter(config)
    if not config.get("silent"): print(docconv)

    docconv.document()
    docconv.update_quarto_yml()

if __name__ == "__main__":
    main()






def demofun_allfine(x = "demo", *args, **kwargs):
    """Demo Function: All fine

    This function is used for testing/demonstrating the
    :py:class:`DocConverter <pyp2qmd.DocConverter.DocConverter>`.
    Docstring all fine, properly includes a title, description, all arguments
    are documented as well as the (dummy) return and exceptions rased.

    Args:
        x (str): Dummy input argument, defaults to \"demo\".
        *args: Takes up all unnamed input arguments, unused.
        **kwargs: Takes up all named arguments, unused.

    Returns:
        str: Simply returns `x` at the end.

    Raises:
        TypeError: If argument `x` is not str.
    """
    if not isinstance(x, str):
        raise TypeError("argument `x` is not str")

    return x


def demofun_wrong(x = "demo", *args, **kwargs):
    """Demo Function: Invalid Docstring

    This function is used for testing/demonstrating the
    :py:class:`DocConverter <pyp2qmd.DocConverter.DocConverter>`.
    Docstring not properly specified, missing documentation for arguments,
    and return.
    """
    if not isinstance(x, str):
        raise TypeError("argument `x` is not str")

    return x


def demofun_wrong2(x = "demo", *args, **kwargs):
    """
    """
    if not isinstance(x, str):
        raise TypeError("argument `x` is not str")

    return x




# -------------------------------------------------------------------
# `setup.py` for the small packages "python package to quarto
# documentation" package.
# -------------------------------------------------------------------

from setuptools import setup, find_namespace_packages

ISRELEASED  = False
VERSION     = "0.1.0"

setup(
    name         = "pyp2qmd",
    version      = VERSION,
    author       = "Reto Stauffer",
    author_email = "reto.stauffer@uibk.ac.at",
    description  = "Parses docstrings of all exported classes and methods of a python package and turns them into quarto markdown (qmd) files. Alongside, a basic structure for a quarto website is created, intended to easily document your python package with quarto.",
    long_description = open("README.md").read(),
    long_description_content_type = "text/markdown",
    url = "https://github.com/retostauffer/pyp2qmd",
    project_urls = {
        "GitHub Issues": "https://github.com/retostauffer/pyp2qmd/issues"
    },
    classifiers = [
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent"
    ],
    install_requires     = ["docstring_parser", "pyyaml"],
    packages             = find_namespace_packages("src"),
    package_dir          = {"": "src"},
    python_requires      = ">=3.8",
    zip_safe             = False,

    # Static files (templates)
    include_package_data = True,
    package_data = {"pyp2qmd": ["templates/*"]},

    # Executables
    entry_points = {
        "console_scripts": [
            "pyp2qmd = pyp2qmd.bin.pyp2qmd:main"
        ]
    }

)
